/* ---------------------------------------------------------------------------
 * Haskell interface to Maple
 * 
 * (c) 2000
 *     Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
 *     Hans-Wolfgang Loidl <hwloidl@cee.hw.ac.uk>
 *
 * $Id: maple.c,v 1.1 2000/11/08 14:13:48 schreine Exp $
 * -------------------------------------------------------------------------*/

/* version string used for communication with Maple, see start file */
#define VERSION "HaskellMaple 1.0"

/* environment variable holding name of root directory */
#define MAPLEENV "MAPLEROOT"

/* name of maple binary */
#define MAPLEBIN "maple"

/* comma separated list of arguments to maple binary */
#define MAPLEARGS "-q"

/* name of file holding maple initialization commands */
#define MAPLEINIT "init.maple"

/* name of file holding maple startup commands */
#define MAPLESTART "start.maple"

/* name of file for error messages appended by process identifier */
#define DEBUGFILE "error"

/* length of line read in one piece */
#define LINELEN 160

/* initial length of buffers, will be doubled on demand */
#define BUFFERLEN 16000

/* ----------------------------------------------------------------------------
 * Posix interface
 * ------------------------------------------------------------------------- */

/* required for select() feature */
#define __USE_BSD

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/time.h>

/* ----------------------------------------------------------------------------
 * exported function interfaces
 * ------------------------------------------------------------------------- */

#include "maple.h"

/* used by Haskell for locking interface */
int lockVar;

/* ----------------------------------------------------------------------------
 * private variables
 * ------------------------------------------------------------------------- */

/* id of Maple process */
static pid_t pid = (pid_t)NULL;

/* communication pipes from/to Maple process */
int in[2], out[2];
static FILE *inpipe = NULL;
static FILE *outpipe = NULL;

/* buffer for reading data from maple pipe */
static char *inbuffer = NULL;
static int inlen = 0;

/* original signal handlers overridden by our handler */
static struct sigaction termHandler;
static struct sigaction intHandler;
static struct sigaction quitHandler;

/* statically allocated temp variables */
MapleRet       mapleRetCore;
MapleRetInt    mapleRetInt;
MapleRetChar   mapleRetChar;
MapleRetString mapleRetString;

/* ----------------------------------------------------------------------------
 *
 * auxiliary functions
 *
 * ------------------------------------------------------------------------- */

/* ----------------------------------------------------------------------------
 * return absolute path for filename
 * ------------------------------------------------------------------------- */
static char *filePath(char *filename)
{
  char* root;
  char* path;
  int rootlen;
  int n;
  root = getenv(MAPLEENV);
  if (root == NULL) root = ".";
  rootlen = strlen(root);
  n = rootlen+1+strlen(filename)+1;
  path = (char*)malloc(n);
  if (path == NULL)
  {
    fprintf(stderr, "maple interface error: cannot allocate path buffer\n");
    exit(-1);
  }
  strcpy(path, root);
  path[rootlen] = '/';
  strcpy(path+rootlen+1, filename);
  return path;
}

/* ----------------------------------------------------------------------------
 * print message to standard error and to debug file
 * ------------------------------------------------------------------------- */
static void report(char *key, char *info)
{
  int pathlen;
  char *debug;
  char *path;
  FILE *file;
  fprintf(stderr, "maple interface error: %s (%s)\n", key, info);
  debug = filePath(DEBUGFILE);
  pathlen = strlen(debug);
  path = (char*)malloc(pathlen+20);
  if (path == NULL) return;
  strcpy(path, debug);
  path[pathlen] = '.';
  sprintf(path+pathlen+1, "%d", getpid());
  file = fopen(path, "a");
  if (file != NULL) 
  {    
    fprintf(file, "maple interface error: %s (%s)\n", key, info);
    fclose(file);
  }
}

/* ----------------------------------------------------------------------------
 * abort with an error message printed to standard error and to debug file
 * ------------------------------------------------------------------------- */
static void error(char *key, char *info)
{
  report(key, info);
  exit(-1);
}

/* ----------------------------------------------------------------------------
 * allocate n bytes of memory, abort if not possible
 * ------------------------------------------------------------------------- */
static void *alloc(size_t n)
{
  void *buffer = malloc(n);
  if (buffer == NULL) 
    {
      char string[20];
      sprintf(string, "%d", n);
      error("cannot allocate buffer", string);
    }
  return buffer;
}

/* ----------------------------------------------------------------------------
 * return next line from maple process, uses inbuffer to hold data
 * ------------------------------------------------------------------------- */
static char *readMaple()
{
  char *result;
  char *newbuffer;
  int newlen;
  int n, end;
  end = 0;
  while (1)
  {
    result = fgets(inbuffer+end, inlen-end, inpipe);
    if (result == NULL) 
      error("end of input encountered", "readMaple");
    n = strlen(inbuffer);
    if (n < inlen-1) 
    {
      /* strip trailing newline */
      if (n > 0 && inbuffer[n-1] == '\n')
	inbuffer[n-1] = '\0';      
      return inbuffer;
    }
    newlen = 2*inlen;
    newbuffer = (char*)alloc(newlen);
    strcpy(newbuffer, inbuffer);
    free(inbuffer);
    end = inlen-1;
    inlen = newlen;
    inbuffer = newbuffer;
  }
}

/* ----------------------------------------------------------------------------
 * read next line from maple process and convert to integer
 * ------------------------------------------------------------------------- */
int readMapleCode()
{
  int i;
  char *line;
  line = readMaple();
  errno = 0;
  i = strtol(line, 0, 10);
  if (errno != 0) error("cannot convert maple return code", line);
  return i;
}

/* ----------------------------------------------------------------------------
 *
 * process initialization and termination
 *
 * ------------------------------------------------------------------------- */

/* ----------------------------------------------------------------------------
 * signal handler terminates Maple process and invokes original handler
 * ------------------------------------------------------------------------- */
void signalHandler(int number)
{
  char line[LINELEN];
  sprintf(line, "%d", number);
  report("signal received", line);
  mapleTerm();
  sigaction (SIGTERM, &termHandler, NULL);
  sigaction (SIGINT, &intHandler, NULL);
  sigaction (SIGQUIT, &quitHandler, NULL);
  raise(number);
}

/* ----------------------------------------------------------------------------
 * install handler for interrupting signals
 * ------------------------------------------------------------------------- */
void installHandler (void)
{
  struct sigaction action;
  sigset_t block_mask;
  int result;

  sigemptyset (&block_mask);
  action.sa_handler = signalHandler;
  action.sa_mask = block_mask;
  action.sa_flags = 0;

  result = sigaction (SIGTERM, &action, &termHandler);
  if (result < 0) error("cannot install signal handler", "SIGTERM");
  sigaction (SIGINT, &action, &intHandler);
  if (result < 0) error("cannot install signal handler", "SIGINT");
  sigaction (SIGQUIT, &action, &quitHandler);
  if (result < 0) error("cannot install signal handler", "SIGQUIT");
}

/* ----------------------------------------------------------------------------
 * start and initialize maple process, abort if error
 * ------------------------------------------------------------------------- */
void mapleInit(void)
{
  int result;
  char *path;

  /* check startup */
  if (pid != 0) error("process already initialized", "mapleInit");

  /* install signal handler to terminate Maple process properly */
  installHandler();

  /* setup communication */
  result = pipe(in);
  if (result < 0) error("cannot open pipe", "in");
  result = pipe(out);
  if (result < 0) error("cannot open pipe", "out");

  /* fork process */
  pid = fork();
  if (pid < 0) 
  {
    char string[20];
    sprintf(string, "%d", pid);
    error("cannot create process", string);
  }

  /* redirect in subprocess stdin and stdout to pipes */
  if (pid == 0)
  {
    int result;
    result = dup2(out[0], STDIN_FILENO);
    if (result == -1) error("cannot duplicate file descriptor", "stdin");
    close(out[0]);
    result = dup2(in[1], STDOUT_FILENO);
    if (result == -1) error("cannot duplicate file descriptor", "stdout");
    close(in[1]);
    path = filePath(MAPLEBIN);
    execl(path, MAPLEBIN, MAPLEARGS, NULL);
    error("cannot execute file", path);
  }

  /* check that child lives */
  sleep(1);
  if (waitpid (pid, NULL, WNOHANG) != 0)
    error("child process has terminated", "mapleInit");

  /* open in main process buffered pipes to subprocess*/
  inpipe = fdopen(in[0], "r");
  if (inpipe == NULL) error("cannot open buffered pipe", "in");
  outpipe = fdopen(out[1], "w");
  if (outpipe == NULL) error("cannot open buffered pipe", "out");

  /* input is not buffered for MapleResultAvailable() */
  setvbuf(inpipe, NULL, _IONBF, 0);

  /* allocate input buffer for communication from subprocess */
  inbuffer = (char*)alloc(BUFFERLEN);
  inlen = BUFFERLEN;

  /* send initialization and startup commands to subprocess */
  fprintf(outpipe, "read `%s`:\n", filePath(MAPLEINIT));
  fprintf(outpipe, "read `%s`:\n", filePath(MAPLESTART));
  fflush(outpipe);

  /* make termination sure */
  atexit(mapleTerm);

  /* wait for process to be ready */
  readMaple();

  /* first line in protocol is name of debug file */
  fprintf(outpipe, "%s\n", filePath(DEBUGFILE));
  fflush(outpipe);
}

/* ----------------------------------------------------------------------------
 * terminate Maple process before handling signal in the original way
 * ------------------------------------------------------------------------- */
void mapleTerm(void)
{
  if (inbuffer != NULL) 
    {
      free(inbuffer);
      inbuffer = NULL;
    }
  if (inpipe != NULL)   
  {
    fclose(inpipe);
    inpipe = NULL;
  }
  if (outpipe != NULL)  
    {
      fprintf(outpipe, "%s\n", "terminate");
      fclose(outpipe);
      outpipe = NULL;
    }
  if (pid != 0) 
    {
      sleep(1);
      kill(pid, SIGTERM);
      kill(pid, SIGKILL);
      pid = 0;
    }
}

/* ----------------------------------------------------------------------------
 *
 * external interface
 *
 * ------------------------------------------------------------------------- */

/* ----------------------------------------------------------------------------
 * construct return value from C-object
 * cobj needs to be copied!
 * ------------------------------------------------------------------------- */
static MapleRet *mapleRet(int ret, char* cobj)
{
  int n = strlen(cobj);
  MapleRet *m = &mapleRetCore;
  m->ret = ret;
  m->val.n = n;
  m->val.addr = cobj;
  return m;
}

/* ----------------------------------------------------------------------------
 * convert Haskell int to Maple integer
 * ------------------------------------------------------------------------- */
MapleRet *mapleInt(int i)
{
  char* cobj;
  int ret;
  fprintf(outpipe, "in\n%d\n", i);
  fflush(outpipe);
  ret = readMapleCode();
  cobj = readMaple();
  if (ret != 0) report("error in converting Haskell integer", cobj);
  return mapleRet(ret, cobj);
}

/* ----------------------------------------------------------------------------
 * convert Maple integer to Haskell Integer
 * ------------------------------------------------------------------------- */
MapleRet *mapleInteger(char *i, int n)
{
  char* cobj;
  int ret;
  fprintf(outpipe, "in\n");
  writeString(outpipe, i, n);
  fprintf(outpipe, "\n");
  fflush(outpipe);
  ret = readMapleCode();
  cobj = readMaple();
  if (ret != 0) report("error in converting Haskell integer", cobj);
  return mapleRet(ret, cobj);
}

/* ----------------------------------------------------------------------------
 * convert Maple integer to Haskell int
 * ------------------------------------------------------------------------- */
MapleRetInt *haskellInt(char* hobj, int n)
{
  MapleRetInt *m = &mapleRetInt;
  char *str;
  long int l;
  fprintf(outpipe, "out\n");
  writeLine(outpipe, hobj, n);
  fflush(outpipe);
  m->ret = readMapleCode();
  str = readMaple();
  if (m-> ret != 0) 
  {
    report("error in converting Maple integer", str);
    return m;
  }
  errno = 0;
  l = strtol(str, NULL, 10);
  if (errno != 0 || l < INT_MIN || l > INT_MAX)
  {
    m->ret = -1;
    report("error in converting Maple integer", str);
  }
  else
    m->val = l;
  return m;
}

/* ----------------------------------------------------------------------------
 * convert Haskell char to Maple char
 * ------------------------------------------------------------------------- */
MapleRet *mapleChar(char c)
{
  char* cobj;
  int ret;
  fprintf(outpipe, "in\n'%c'\n", c);
  fflush(outpipe);
  ret = readMapleCode();
  cobj = readMaple();
  if (ret != 0) report("error in converting Haskell character", cobj);
  return mapleRet(ret, cobj);
}

/* ----------------------------------------------------------------------------
 * convert Maple char to Haskell char
 * ------------------------------------------------------------------------- */
MapleRetChar *haskellChar(char *hobj, int n)
{
  MapleRetChar *m = &mapleRetChar;
  char *str;
  fprintf(outpipe, "out\n");
  writeLine(outpipe, hobj, n);
  fflush(outpipe);
  m->ret = readMapleCode();
  str = readMaple();
  if (m->ret != 0) 
  {
    report("error in converting Maple character", str);
    return m;
  }
  if (strlen(str) != 1)
  {
    report("error in converting Maple character", str);
    m->ret = -1;
  }
  else
    m->val = str[0];
  return m;
}

/* ----------------------------------------------------------------------------
 * convert Maple string to Haskell string
 * ------------------------------------------------------------------------- */
MapleRetString *haskellString(char* hobj, int n)
{
  MapleRetString *m = &mapleRetString;
  char *str;
  fprintf(outpipe, "out\n");
  writeLine(outpipe, hobj, n);
  fflush(outpipe);
  m->ret = readMapleCode();
  str = readMaple();
  if (m->ret != 0) 
  {
    report("error in converting Maple string", str);
    return m;
  }
  m->len = strlen(str);
  m->str = str;
  return m;
}

/* ----------------------------------------------------------------------------
 * convert Haskell string to Maple expression
 * ------------------------------------------------------------------------- */
MapleRet *mapleExpr(char *s, int n)
{
  char* cobj;
  int ret;
  fprintf(outpipe, "in\n");
  writeLine(outpipe, s, n);
  fflush(outpipe);
  ret = readMapleCode();
  cobj = readMaple();
  if (ret != 0) report("error in converting Haskell expression", cobj);
  return mapleRet(ret, cobj);
}

/* ----------------------------------------------------------------------------
 * start evaluation with n arguments and single result
 * ------------------------------------------------------------------------- */
void mapleStartEval(int n)
{
  fprintf(outpipe, "fun\n");
  fprintf(outpipe, "%d\n", n);
}

/* ----------------------------------------------------------------------------
 * return handle of pipe to Maple
 * ------------------------------------------------------------------------- */
FILE *mapleOut()
{
  return outpipe;
}

/* ----------------------------------------------------------------------------
 * end evaluation
 * ------------------------------------------------------------------------- */
void mapleEndEval()
{
  fflush(outpipe);
}

/* ----------------------------------------------------------------------------
 * returns non-zero iff result is available 
 * ------------------------------------------------------------------------- */
int mapleResultAvailable()
{
  static fd_set inset;
  static struct timeval timeout;
  int r;
  FD_ZERO(&inset);
  FD_SET(in[0], &inset);
  timeout.tv_sec = 0;
  timeout.tv_usec = 10000;
  r = select(FD_SETSIZE, &inset, NULL, NULL, &timeout);
  if (r == -1) error("error in select", "mapleResultAvailable");
  return r != 0;
}

/* ----------------------------------------------------------------------------
 * read result from Maple pipe
 * ------------------------------------------------------------------------- */
MapleRet *mapleResult()
{
  int ret = readMapleCode();
  char* cobj = readMaple();
  if (ret != 0) report("error in Maple function", cobj);
  return mapleRet(ret, cobj);
}

/* ----------------------------------------------------------------------------
 * start evaluation with n arguments and sequence of results
 * ------------------------------------------------------------------------- */
void mapleStartEvalN(int n)
{
  fprintf(outpipe, "lst\n");
  fprintf(outpipe, "%d\n", n);
}

/* ----------------------------------------------------------------------------
 * read number of results from Maple pipe
 * ------------------------------------------------------------------------- */
int mapleResultN()
{
  int ret = readMapleCode();
  if (ret < 0) 
    {
      char* cobj = readMaple();
      report("error in Maple function", cobj);
    }
  return ret;
}

/* ----------------------------------------------------------------------------
 * read next result from Maple pipe
 * ------------------------------------------------------------------------- */
MapleRet *mapleNextResult()
{
  char* cobj = readMaple();
  return mapleRet(0, cobj);
}

/* ----------------------------------------------------------------------------
 *
 * Haskell helper functions
 *
 * ------------------------------------------------------------------------- */

/* ----------------------------------------------------------------------------
 * print string starting at addr 
 * ------------------------------------------------------------------------- */
void writeString(FILE *fd, char *addr, int n) 
{
  int ret;
  ret = fwrite(addr, 1, n, fd);
  if (ret != n) 
    error("error in writing", "string");
}

/* ----------------------------------------------------------------------------
 * print n bytes starting at addr 
 * ------------------------------------------------------------------------- */
void writeLine(FILE *fd, char *addr, int n) 
{
  int ret;

  /* Maple reads only up to 1022 chars per line */
  /* VERSION key helps to distinguish lines whose length is multiple of this */
  if (n % 1022 == 0)
    writeString(fd, VERSION, strlen(VERSION));

  writeString(fd, addr, n);
  ret = fputc('\n', fd);
  if (ret == EOF) 
    error("error in writing", "newline");
}

/* ----------------------------------------------------------------------------
 *
 * debugging interface
 *
 * ------------------------------------------------------------------------- */

/* ----------------------------------------------------------------------------
 * can be used as main program
 * ------------------------------------------------------------------------- */
int mapleMain(int argc, char* argv[])
{
  int n, m;
  char line[LINELEN];
  char *result;
  int ret;

  /* initialize Maple */
  mapleInit();

  printf("ready\n");

  /* receive input from stdin */
  result = fgets(line, LINELEN, stdin);
  while (result != NULL)
  {
    int l = strtol(result, NULL, 10);
    int i;
    
    for (i = 0; i < l; i++)
    {
      result = fgets(line, LINELEN, stdin);
      n = strlen(result);

      /* send input to Maple */
      m = fwrite(line, 1, n, outpipe);
      if (m < n) error("cannot write to output pipe", "");
    }
    fflush(outpipe);

    /* receive answer from Maple */
    ret = readMapleCode();
    result = readMaple();

    /* write answer to stdout */
    printf("%d\n", ret);
    fputs(result, stdout);

    /* receive input from stdin */
    result = fgets(line, LINELEN, stdin);
  }
  
  /* terminate Maple */
  mapleTerm();
  return 0;
}

/* ---------------------------------------------------------------------------
 * $Id: maple.c,v 1.1 2000/11/08 14:13:48 schreine Exp $
 * -------------------------------------------------------------------------*/
